#include "header\gba.h"
#include "header\bg.h"
#include "header\timer.h"
#include "header\dma.h"
#include "header\irq.h"
#include "header\sound.h"
#include "header\sprite.h"
#include "header\blend.h"

//#include "header\font\BGFont.h"
//#include "header\font\BitMapFont.h"

#include "lib.h"
#include "syscall.h"
#include "resource.h"

#include <stdlib.h>
#include <math.h>

//Mode3

//---------------------------------------------------------------------------
#define STATE_INIT 				1		//is
#define STATE_RESET				2
#define STATE_READY				3
#define STATE_SHOOT				4

#define PI						(f32)3.1415926
#define ZERO1					(f32)1e-12
#define BOUND_E					(f32)0.8;	//W

#define MODE3_COLOR_WHITE		RGB(31,31,31)
#define MODE3_COLOR_BLOCK		RGB( 0, 0, 0)
#define MODE3_COLOR_RED			RGB(31, 0, 0)

//---------------------------------------------------------------------------
typedef struct {
	f32 x;
	f32 y;
} ST_POINT;

typedef struct {
	f32 x;
	f32 y;
	f32 r;
} ST_CIRCLE;

typedef struct {
	f32 x1;
	f32 y1;
	f32 x2;
	f32 y2;
} ST_LINE;

//---------------------------------------------------------------------------
//lib.c
extern ST_TIMER Timer;

//---------------------------------------------------------------------------
ST_CIRCLE Circle;
ST_LINE   Line;

s16 Angle;
f32 SpeedF;
f32 SpeedX;
f32 SpeedY;

u16 Key;
u8  State;						//is󋵃tO
//---------------------------------------------------------------------------
void StateInit(); 
void StateReset();
void StateReady(); 
void StateShoot();

void InitWS();
void InitKey();
void IRQUserHandler();

void DrawAll();
void DrawCircle(ST_CIRCLE c);
void DrawLine(u8 sx, u8 sy, u8 ex, u8 ey, u16 col);

f32  GetRadian(f32 deg);

//---------------------------------------------------------------------------
int main()
{
	State = STATE_INIT;

	for(;;)
	{
		switch(State)
		{
			case STATE_INIT:  StateInit();  break;
			case STATE_RESET: StateReset(); break;
			case STATE_READY: StateReady(); break;
			case STATE_SHOOT: StateShoot(); break;

			default:
				DebugMessage((u8*)"Err: Main State");
				break;
		}
	}

	return 0;
}
//---------------------------------------------------------------------------
void StateInit()
{
	InitWS();
	InitKey();

	SetMode(MODE_3 | BG2_ENABLE | OBJ_ENABLE | OBJ_MAP_2D);

	State = STATE_RESET;
}
//---------------------------------------------------------------------------
void InitWS()
{
	REG_WSCNT = CST_ROM0_1ST_3WAIT | CST_ROM0_2ND_1WAIT;
}
//---------------------------------------------------------------------------
void InitKey()
{
	KeyInit();
	Key = KEY_NONE;

	REG_INTERUPT = (u32)IRQUserHandler;
	REG_IE       = IRQ_BIT_TIMER0;
	REG_IME      = IRQ_MASTER_ON;
	REG_TM0D     = 65536 - 2621;			// (((16*1024*1024) / 64) * 10) / 1000
	REG_TM0CNT   = TM_ENABLE | TM_FREQ_PER_64 | TM_USEIRQ;
}
//---------------------------------------------------------------------------
void IRQUserHandler()
{
	Timer.clockTick += TIMER_INTMS;
	REG_IF = IRQ_BIT_TIMER0;
}
//---------------------------------------------------------------------------
void ScanKey()
{
	if( KeyWaitTimer() == TRUE )
	{
		return;
	}
	if( KeyRepeat() == TRUE )		//keyGet֐ĂԑOɕKsȂ
	{
		return;
	}

	Key = KeyInput();
}
//---------------------------------------------------------------------------
void StateReset()
{
	Angle    = 0;
	SpeedF   = 5;
	SpeedX   =  cos( GetRadian(Angle) ) * SpeedF;
	SpeedY   = -sin( GetRadian(Angle) ) * SpeedF;

	Circle.x = 40;
	Circle.y = 60;
	Circle.r = 8;

	Line.x1  = 140;
	Line.y1  = 50;
	Line.x2  = 50;
	Line.y2  = 150;

//	WaitForVsync();
	DrawAll();

	State = STATE_READY;
}
//---------------------------------------------------------------------------
void StateReady()
{
	ScanKey();

	if(Key & KEY_A)
	{
		State = STATE_SHOOT;
		return;
	}
	if(Key & KEY_B)
	{
		State = STATE_RESET;
		return;
	}

	if( (Key & KEY_LEFT) || (Key & KEY_RIGHT) ||
	    (Key & KEY_UP  ) || (Key & KEY_DOWN ) )
	{
		//pxƑx𒲐܂
		if(Key & KEY_LEFT ) Angle  += 10;
		if(Key & KEY_RIGHT) Angle  -= 10;
//		if(Key & KEY_UP   ) SpeedF += 10;
//		if(Key & KEY_DOWN ) SpeedF -= 10;

		Angle = (Angle + 360) % 360;

		SpeedX =  cos( GetRadian(Angle) ) * SpeedF;
		SpeedY = -sin( GetRadian(Angle) ) * SpeedF;

//		WaitForVsync();
		DrawAll();
	}
}
//---------------------------------------------------------------------------
void StateShoot()
{
	ScanKey();

	if(Key & KEY_B)
	{
		State = STATE_RESET;
		return;
	}

	//x^܂
	Circle.x += SpeedX;
	Circle.y += SpeedY;

	//2̃xNg̐߂
	f32 ax = Circle.x - Line.x1;
	f32 ay = Circle.y - Line.y1;
	f32 bx = Line.x2  - Line.x1;
	f32 by = Line.y2  - Line.y1;

	//ixNgBj̒̓
	f32 inpro = (ax * bx) + (ay * by);
	f32 bl    = (bx * bx) + (by * by);

	//lineƂ̏Փ˔
	if( (inpro > -ZERO1 && inpro < (bl+ZERO1)) == FALSE )
	{
		goto FuncEnd;
	}

	//߂܂
	if( ((ax * ax + ay * ay) - pow(inpro/sqrt(bl), 2) < 8*8) == FALSE )
	{
		goto FuncEnd;
	}

	//Kꂽ@xNg
	f32 nx     =  by / sqrt(bl);
	f32 ny     = -bx / sqrt(bl);
	f32 inpro2 = nx * -SpeedX + ny * -SpeedY;	//

	//@xNg̉
	nx *= inpro2;
	ny *= inpro2;

	SpeedX    = (SpeedX + nx * 2) * BOUND_E;
	SpeedY    = (SpeedY + ny * 2) * BOUND_E;
	Circle.x += SpeedX * 3;
	Circle.y += SpeedY * 3;

FuncEnd:
	DrawAll();
}
//---------------------------------------------------------------------------
void DrawCircle(ST_CIRCLE c)
{
	Mode3DrawFillCircle(c.x, c.y, c.r, MODE3_COLOR_RED);
}
//---------------------------------------------------------------------------
void DrawLine(u8 sx, u8 sy, u8 ex, u8 ey, u16 col)
{
	Mode3DrawLine(sx, sy, ex, ey, col);
}
//---------------------------------------------------------------------------
void DrawAll()
{
	Mode3DrawCls(MODE3_COLOR_BLOCK);

	DrawCircle(Circle);
	Mode3DrawLine(Circle.x, Circle.y, Circle.x + SpeedX*3, Circle.y + SpeedY*3, MODE3_COLOR_WHITE);

	Mode3DrawLine(Line.x1, Line.y1, Line.x2, Line.y2, MODE3_COLOR_WHITE);
}
//---------------------------------------------------------------------------
f32 GetRadian(f32 deg)
{
	return deg * (f32)PI / 180.0;
}
